# cython: embedsignature=True, language_level=3
# distutils: language=c++


from libc.math cimport *
import numpy as np



cdef class CubicSpline:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp2
        cdef double val
        """Gradient of a kernel is given by
        .. math::
            \nabla W = normalization  \frac{dW}{dq} \frac{dq}{dx}
            \nabla W = w_dash  \frac{dq}{dx}

        Here we get `w_dash` by using `dwdq` method
        """
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor ( sigma )
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute sigma * dw_dq
        tmp2 = 2. - q
        if (rij > 1e-12):
            if (q > 2.0):
                val = 0.0
            elif (q > 1.0):
                val = -0.75 * tmp2 * tmp2
            else:
                val = -3.0 * q * (1 - 0.75 * q)
        else:
            val = 0.0

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 2. / 3

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp2
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # kernel and gradient evaluated at q
        tmp2 = 2. - q
        if (q > 2.0):
            w = 0.0
            dw = 0.0

        elif (q > 1.0):
            w = 0.25 * tmp2 * tmp2 * tmp2
            dw = -0.75 * tmp2 * tmp2
        else:
            w = 1 - 1.5 * q * q * (1 - 0.5 * q)
            dw = -3.0 * q * (1 - 0.75 * q)

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp2
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp2 = 2. - q
        if (q > 2.0):
            val = 0.0

        elif (q > 1.0):
            val = 0.25 * tmp2 * tmp2 * tmp2
        else:
            val = 1 - 1.5 * q * q * (1 - 0.5 * q)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class CubicSplineWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public CubicSpline kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class WendlandQuintic:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -5.0 * q * tmp * tmp * tmp

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 0.5

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * (2.0 * q + 1.0)
            dw = -5.0 * q * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * (2.0 * q + 1.0)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class WendlandQuinticWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public WendlandQuintic kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class Gaussian:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        if (q < 3.0):
            if (rij > 1e-12):
                val = -2.0 * q * exp(-q * q)

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        # The inflection point is at q=1/sqrt(2)
        # the deltap values for some standard kernels
        # have been tabulated in sec 3.2 of
        # http://cfd.mace.manchester.ac.uk/sph/SPH_PhDs/2008/crespo_thesis.pdf
        return 0.70710678118654746

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # kernel and gradient evaluated at q
        w = 0.0
        dw = 0.0
        if (q < 3.0):
            w = exp(-q * q)
            dw = -2.0 * q * w

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0
        if (q < 3.0):
            val = exp(-q * q) * fac

        return val

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class GaussianWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public Gaussian kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class QuinticSpline:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp1
        cdef double tmp2
        cdef double tmp3
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp3 = 3. - q
        tmp2 = 2. - q
        tmp1 = 1. - q

        # compute the gradient
        if (rij > 1e-12):
            if (q > 3.0):
                val = 0.0

            elif (q > 2.0):
                val = -5.0 * tmp3 * tmp3 * tmp3 * tmp3

            elif (q > 1.0):
                val = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
                val += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
            else:
                val = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
                val += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
                val -= 75.0 * tmp1 * tmp1 * tmp1 * tmp1
        else:
            val = 0.0

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        # The inflection points for the polynomial are obtained as
        # http://www.wolframalpha.com/input/?i=%28%283-x%29%5E5+-+6*%282-x%29%5E5+%2B+15*%281-x%29%5E5%29%27%27
        # the only permissible value is taken
        return 0.759298480738450

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp1
        cdef double tmp2
        cdef double tmp3
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp3 = 3. - q
        tmp2 = 2. - q
        tmp1 = 1. - q

        # compute the kernel & gradient at q
        if (q > 3.0):
            w = 0.0
            dw = 0.0

        elif (q > 2.0):
            w = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            dw = -5.0 * tmp3 * tmp3 * tmp3 * tmp3

        elif (q > 1.0):
            w = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            w -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2

            dw = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
            dw += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
        else:
            w = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            w -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2
            w += 15. * tmp1 * tmp1 * tmp1 * tmp1 * tmp1

            dw = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
            dw += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
            dw -= 75.0 * tmp1 * tmp1 * tmp1 * tmp1

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp1
        cdef double tmp2
        cdef double tmp3
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp3 = 3. - q
        tmp2 = 2. - q
        tmp1 = 1. - q
        if (q > 3.0):
            val = 0.0

        elif (q > 2.0):
            val = tmp3 * tmp3 * tmp3 * tmp3 * tmp3

        elif (q > 1.0):
            val = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            val -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2

        else:
            val = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            val -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2
            val += 15. * tmp1 * tmp1 * tmp1 * tmp1 * tmp1

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class QuinticSplineWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public QuinticSpline kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class SuperGaussian:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double q2
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        if (q < 3.0):
            if (rij > 1e-12):
                q2 = q * q
                val = q * (2.0 * q2 - self.dim - 4) * exp(-q2)

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        # Found inflection point using sympy.
        if self.dim == 1:
            return 0.584540507426389
        elif self.dim == 2:
            return 0.6021141014644256
        else:
            return 0.615369528365158

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double d
        cdef double fac
        cdef double h1
        cdef double q
        cdef double q2
        cdef double val
        h1 = 1. / h
        q = rij * h1
        d = self.dim

        # get the kernel normalizing factor
        if d == 1:
            fac = self.fac * h1
        elif d == 2:
            fac = self.fac * h1 * h1
        elif d == 3:
            fac = self.fac * h1 * h1 * h1

        # kernel and gradient evaluated at q
        val = 0.0
        if (q < 3.0):
            q2 = q * q
            val = (-d * d * 0.5 + 2.0 * d * q2 - d -
                   2.0 * q2 * q2 + 4 * q2) * exp(-q2)

        return -fac * h1 * val

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double q2
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0
        if (q < 3.0):
            q2 = q * q
            val = exp(-q2) * (1.0 + self.dim * 0.5 - q2) * fac

        return val

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class SuperGaussianWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public SuperGaussian kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class WendlandQuinticC4:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = (-14.0 / 3.0) * q * (1 + 2.5 * q) * \
                    tmp * tmp * tmp * tmp * tmp

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 0.47114274

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * tmp * \
                ((35.0 / 12.0) * q * q + 3.0 * q + 1.0)
            dw = (-14.0 / 3.0) * q * (1 + 2.5 * q) * \
                tmp * tmp * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * tmp * \
                ((35.0 / 12.0) * q * q + 3.0 * q + 1.0)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class WendlandQuinticC4Wrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public WendlandQuinticC4 kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class WendlandQuinticC6:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -5.50 * q * tmp * tmp * tmp * tmp * tmp * \
                    tmp * tmp * (1.0 + 3.5 * q + 4 * q * q)

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 0.4305720757

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (4.0 * q * q * q + 6.25 * q * q + 4.0 * q + 1.0)
            dw = -5.50 * q * tmp * tmp * tmp * tmp * tmp * \
                tmp * tmp * (1.0 + 3.5 * q + 4 * q * q)

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (4.0 * q * q * q + 6.25 * q * q + 4.0 * q + 1.0)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class WendlandQuinticC6Wrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public WendlandQuinticC6 kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class WendlandQuinticC2_1D:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -3.0 * q * tmp * tmp

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 2.0/3

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * (1.5 * q + 1.0)
            dw = -3.0 * q * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * (1.5 * q + 1.0)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class WendlandQuinticC2_1DWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public WendlandQuinticC2_1D kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class WendlandQuinticC4_1D:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -3.5 * q * (2 * q + 1) * tmp * tmp * tmp * tmp

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 0.55195628

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * (2 * q * q + 2.5 * q + 1.0)
            dw = -3.5 * q * (2 * q + 1) * tmp * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * (2 * q * q + 2.5 * q + 1.0)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class WendlandQuinticC4_1DWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public WendlandQuinticC4_1D kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]



cdef class WendlandQuinticC6_1D:
    cdef public long dim
    cdef public double fac
    cdef public double radius_scale
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    cdef inline double dwdq(self, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -0.5 * q * (26.25 * q * q + 27 * q + 9.0) * \
                    tmp * tmp * tmp * tmp * tmp * tmp

        return val * fac

    cpdef double py_dwdq(self, double rij, double h):
        return self.dwdq(rij, h)

    cdef inline double get_deltap(self):
        return 0.47996698

    cpdef double py_get_deltap(self):
        return self.get_deltap()

    cdef inline void gradient(self, double* xij, double rij, double h, double* grad):
        cdef double h1
        cdef double tmp
        cdef double wdash
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    cpdef py_gradient(self, double[:] xij, double rij, double h, double[:] grad):
        self.gradient(&xij[0], rij, h, &grad[0])

    cdef inline double gradient_h(self, double* xij, double rij, double h):
        cdef double dw
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double w
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (2.625 * q * q * q + 4.75 * q * q + 3.5 * q + 1.0)
            dw = -0.5 * q * (26.25 * q * q + 27 * q + 9.0) * \
                tmp * tmp * tmp * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)

    cpdef double py_gradient_h(self, double[:] xij, double rij, double h):
        return self.gradient_h(&xij[0], rij, h)

    cdef inline double kernel(self, double* xij, double rij, double h):
        cdef double fac
        cdef double h1
        cdef double q
        cdef double tmp
        cdef double val
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (2.625 * q * q * q + 4.75 * q * q + 3.5 * q + 1.0)

        return val * fac

    cpdef double py_kernel(self, double[:] xij, double rij, double h):
        return self.kernel(&xij[0], rij, h)



cdef class WendlandQuinticC6_1DWrapper:
    """Reasonably high-performance convenience wrapper for Kernels.
    """

    cdef public WendlandQuinticC6_1D kern
    cdef double[3] xij, grad
    cdef public double radius_scale
    cdef public double fac

    def __init__(self, kern):
        self.kern = kern
        self.radius_scale = kern.radius_scale
        self.fac = kern.fac

    cpdef double kernel(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        return self.kern.kernel(xij, rij, h)

    cpdef gradient(self, double xi, double yi, double zi, double xj, double yj, double zj, double h):
        cdef double* xij = self.xij
        xij[0] = xi-xj
        xij[1] = yi-yj
        xij[2] = zi-zj
        cdef double rij = sqrt(xij[0]*xij[0] + xij[1]*xij[1] +xij[2]*xij[2])
        cdef double* grad = self.grad
        self.kern.gradient(xij, rij, h, grad)
        return grad[0], grad[1], grad[2]




""" Thacker Basin Test Case (Oscillation of a fluid column in a frictionless
paraboloid basin) (4.4 hr)

The case is described in "Accurate particle splitting for smoothed particle
hydrodynamics in shallow water with shock capturing", R. Vacondio, B. D. Rogers
and P.K. Stansby, Int. J. Numer. Meth. Fluids, Vol 69, pp 1377-1410 (2012).
DOI: 10.1002/fld.2646

"""
import os

# Numpy
from numpy import (ones_like, zeros, zeros_like, mgrid, pi, array,
                   arange, sqrt, concatenate, sin, cos, savez)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration, FluidBottomElevation, FluidBottomGradient,
        FluidBottomCurvature, GradientCorrectionPreStep, GradientCorrection,
        BedGradient, BedCurvature
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator

# PySPH Interpolator
from pysph.tools.interpolator import Interpolator


# Constants
rho_w = 1000.0
g = 9.81
dim = 2.0


class ThackerBasin(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.5,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--r", action="store", type=float, dest="r", default=3000.0,
            help="Initial radius of the fluid column (m)")
        group.add_argument(
            "--n", action="store", type=float, dest="n", default=75,
            help="Number of concentric fluid particle circles (Determines\
            spacing btw particles, dr = r/n)")
        group.add_argument(
            "--x_cen_fluid", action="store", type=float, dest="x_cen_fluid",
            default=1500.0, help="x-coordinate of center of the circular fluid\
            column (m)")
        group.add_argument(
            "--y_cen_fluid", action="store", type=float, dest="y_cen_fluid",
            default=0.0, help="y-coordinate of center of the circular fluid\
            column (m)")
        group.add_argument(
            "--zo", action="store", type=float, dest="zo", default=10.0,
            help="Distance of parabola origin from bottom (m)")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.r = self.options.r
        self.n = self.options.n
        self.x_cen_fluid = self.options.x_cen_fluid
        self.y_cen_fluid = self.options.y_cen_fluid
        self.zo = self.options.zo
        self.omega = sqrt(2*g*self.zo) / self.r  # Angular Frequency

    def create_particles(self):
        n = self.n
        hdx = self.hdx
        fluid_rad = self.r
        dr = (fluid_rad-100) / n
        zo = self.zo

        # Bed
        dxb = 50.
        xb, yb = mgrid[-5000:5000:dxb, -5000:5000:dxb]
        b = zo * ((xb**2 + yb**2)/fluid_rad**2)

        Vb = ones_like(xb) * dxb * dxb
        hb = ones_like(xb) * hdx * dxb

        bed = gpa_swe(name='bed', x=xb, y=yb, V=Vb, b=b, h=hb)

        # For gradient correction
        len_b = len(bed.x) * 9
        bed.add_constant('m_mat', [0.0] * len_b)

        # Fluid
        x = zeros(0)
        y = zeros(0)

        # Create circular patch in a radial grid
        rad = 0.0
        for j in range(1, n+1):
                npnts = 4 * j
                dtheta = (2*pi) / npnts

                theta = arange(0, 2*pi-1e-10, dtheta)
                rad = rad + dr

                _x = rad * cos(theta)
                _y = rad * sin(theta)

                x = concatenate((x, _x))
                y = concatenate((y, _y))

        x += self.x_cen_fluid
        y += self.y_cen_fluid

        h = ones_like(x) * hdx * dr
        h0 = ones_like(x) * hdx * dr

        # Distance between fluid center and bed center along x-direction
        zeta = self.x_cen_fluid - 0.0

        u = zeros_like(x)
        v = ones_like(x) * -(zeta*self.omega)
        vh = ones_like(x) * -(zeta*self.omega)

        # At t = 0.0s
        fluid_surf_hei = zo + (2*zeta*(zo/fluid_rad)
                               * ((x/fluid_rad) - (zeta/(2.0*fluid_rad))))

        fluid = gpa_swe(x=x, y=y, h=h, h0=h0, u=u, v=v, vh=vh, name='fluid')

        compute_fluid_elevation([fluid, bed])

        dw = fluid_surf_hei - fluid.b

        rho = dw * rho_w
        rho0 = dw * rho_w
        m = rho * (1.56*dr*dr)

        fluid.m = m
        fluid.rho = rho
        fluid.rho0 = rho0
        fluid.dw = dw

        compute_initial_props([fluid])

        return [fluid, bed]

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = (2*pi) / self.omega
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.3,
            adaptive_timestep=True,
            output_at_times=[0.015*tf, 0.505*tf, 0.980*tf],
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(
                                dest='fluid', sources=['fluid']
                                ),
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid']),
                    ]
                ),
            Group(
                equations=[
                    SWEOS(dest='fluid'),
                    ]
                ),
            Group(
                equations=[
                    FluidBottomElevation(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomGradient(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomCurvature(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(dim=dim, dest='fluid',
                                         sources=['fluid'])
                    ],
                ),
            ]
        return equations

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth_and_vel()

    def _plot_depth_and_vel(self):
        from matplotlib import pyplot as plt
        t_arr = []
        u_arr = []
        dw_arr = []
        v_arr = []

        # Properties of the fluid at this location are found at various times
        # Note: This location is wetted at all times by the fluid
        x_loc_to_interpolate = 0.0
        y_loc_to_interpolate = 0.0

        kernel = CubicSpline(dim=2)
        for fname in self.output_files:
            data = load(fname)
            fluid = data['arrays']['fluid']
            t_arr.append(data['solver_data']['t'])
            interp = Interpolator([fluid], kernel=kernel)
            interp.set_interpolation_points(x_loc_to_interpolate,
                                            y_loc_to_interpolate)
            u_interp = interp.interpolate('u')
            v_interp = interp.interpolate('v')
            dw_interp = interp.interpolate('dw')
            u_arr.append(u_interp.item())
            v_arr.append(v_interp.item())
            dw_arr.append(dw_interp.item())

        t = array(t_arr)
        x = zeros_like(t)
        y = zeros_like(t)
        fluid_rad = self.r
        zo = self.zo
        zeta = self.x_cen_fluid - 0.0
        omega = self.omega

        def actual_fluid_surf_hei(x, y, t):
            fluid_surf_hei = zo + (2*zeta*(zo/fluid_rad)
                                   * ((x/fluid_rad)*cos(omega*t)
                                   - (y/fluid_rad)*sin(omega*t)
                                   - (zeta/(2.0*fluid_rad))))
            return fluid_surf_hei

        def fluid_bottom_hei(x, y):
            return zo * ((x**2+y**2)/fluid_rad**2)

        def u_actual_func(t):
            return -zeta*omega*sin(omega*t)

        def v_actual_func(t):
            return -zeta*omega*cos(omega*t)

        dw_actual = actual_fluid_surf_hei(x, y, t) - fluid_bottom_hei(x, y)
        u_actual = u_actual_func(t)
        v_actual = v_actual_func(t)

        # Time period
        T = (2*pi) / omega

        if len(dw_arr) == 0 or len(u_arr) == 0 or len(v_arr) == 0:
            return

        dw_min = min(dw_arr)
        dw_max = max(dw_arr)
        u_min = min(u_arr)
        u_max = max(u_arr)
        v_min = min(v_arr)
        v_max = max(v_arr)

        b = fluid_bottom_hei(x, y)

        fname_res = os.path.join(self.output_dir, 'results.npz')
        savez(
            fname_res, t=t_arr, dw_num=dw_arr, dw_actual=dw_actual,
            u_num=u_arr, u_actual=u_actual, v_num=v_arr, v_actual=v_actual,
            x=x, y=y, b=b
            )

        plt.clf()
        plt.plot(t_arr/T, dw_actual, 'r-', label='Analytical Solution')
        plt.plot(t_arr/T, dw_arr, 'bo', markersize=2, label='SWE-SPH')
        plt.legend()
        plt.ylim(0.8*dw_min, 1.5*dw_max)
        plt.xlim(0, 1)
        plt.xlabel('t/T')
        plt.ylabel('dw (m)')
        fig = os.path.join(self.output_dir, "depth")
        plt.savefig(fig, dpi=300)

        plt.clf()
        plt.plot(t_arr/T, u_actual, 'ro', markersize=2,
                 label='Analytical Solution')
        plt.plot(t_arr/T, u_arr, 'bo', markersize=2, label='SWE-SPH')
        plt.legend()
        plt.ylim(1.5*u_min, 1.5*u_max)
        plt.xlim(0, 1)
        plt.xlabel('t/T')
        plt.ylabel('u (m/s)')
        fig = os.path.join(self.output_dir, "velocity1")
        plt.savefig(fig, dpi=300)

        plt.clf()
        plt.plot(t_arr/T, v_actual, 'ro', markersize=2,
                 label='Analytical Solution')
        plt.plot(t_arr/T, v_arr, 'bo', markersize=2, label='SWE-SPH')
        plt.legend()
        plt.ylim(1.5*v_min, 1.5*v_max)
        plt.xlim(0, 1)
        plt.xlabel('t/T')
        plt.ylabel('v (m/s)')
        fig = os.path.join(self.output_dir, "velocity2")
        plt.savefig(fig, dpi=300)


def compute_initial_props(particles):
    one_time_equations = [
        Group(
            equations=[
                SWEOS(dest='fluid'),
                ],
            )
    ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


def compute_fluid_elevation(particles):
    one_time_equations = [
            Group(
                equations=[
                    FluidBottomElevation(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    GradientCorrectionPreStep(dest='bed', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    GradientCorrection(dest='bed', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    BedGradient(dest='bed', sources=['bed']),
                    ]
                ),
            Group(
                equations=[
                    BedCurvature(dest='bed', sources=['bed']),
                    ]
                ),
        ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = ThackerBasin()
    app.run()
    app.post_process(app.info_filename)




import os
import shutil
import tempfile
import unittest

import numpy as np

from pysph.solver.utils import load, get_files
from pysph.tools import run_parallel_script

MY_DIR = run_parallel_script.get_directory(__file__)


def get_example_script(script):
    """Given a relative posix path to a script located inside the
    pysph.examples package, return the full path to it.

    """
    ex_dir = os.path.join(os.path.dirname(os.path.dirname(MY_DIR)), 'examples')
    return os.path.join(ex_dir, *script.split('/'))


class ExampleTestCase(unittest.TestCase):
    """ A script to run an example in serial and parallel and compare results.

    The comparison is performed in the _test function and currently just checks
    the positions of the particles and tests if they are close.

    To test an example in parallel, subclass from ExampleTest and
    write a test function like so:

    def test_elliptical_drop(self):
        dt = 1e-5; tf = 100*dt
        serial_kwargs = dict(timestep=dt, tf=tf, ghost_layers=1, lb_freq=5)
        self.run_example(
            'dambreak3D.py', nprocs=4, timeout=900,
            serial_kwargs=serial_kwargs
        )

    """
    def _kwargs_to_command_line(self, kwargs):
        """Convert a dictionary of keyword arguments to a list of command-line
        options.
        """
        def _key_to_option(arg):
            return arg.replace('_', '-')

        cmd_line = []
        for key, value in kwargs.items():
            option = _key_to_option(key)
            if value is None:
                arg = "--{option}".format(option=option)
            else:
                arg = "--{option}={value}".format(
                    option=option, value=str(value)
                )

            cmd_line.append(arg)
        return cmd_line

    def run_example(self, filename, nprocs=2, timeout=300, atol=1e-14,
                    serial_kwargs=None, extra_parallel_kwargs=None):
        """Run an example and compare the results in serial and parallel.

        Parameters:
        -----------

        filename : str
            The name of the file to run

        nprocs : int
            Number of processors to use for the parallel run.

        timeout : float
            Time in seconds to wait for execution before an error is raised.

        atol: float
            Absolute tolerance for differences between the runs.

        serial_kwargs : dict
            The options to pass for a serial run.  Note that if the value of a
            particular key is None, the option is simply set and no value
            passed.  For example if `openmp=None`, then `--openmp` is used.

        extra_parallel_kwargs: dict
            The extra options to pass for the parallel run.

        """
        if serial_kwargs is None:
            serial_kwargs = {}
        if extra_parallel_kwargs is None:
            extra_parallel_kwargs = {}

        parallel_kwargs = dict(serial_kwargs)
        parallel_kwargs.update(extra_parallel_kwargs)

        prefix = os.path.splitext(os.path.basename(filename))[0]
        # dir1 is for the serial run
        dir1 = tempfile.mkdtemp()
        serial_kwargs.update(fname=prefix, directory=dir1)

        # dir2 is for the parallel run
        dir2 = tempfile.mkdtemp()
        parallel_kwargs.update(fname=prefix, directory=dir2)

        serial_args = self._kwargs_to_command_line(serial_kwargs)
        parallel_args = self._kwargs_to_command_line(parallel_kwargs)

        try:
            # run the example script in serial
            run_parallel_script.run(
                filename=filename, args=serial_args, nprocs=1,
                timeout=timeout, path=MY_DIR
            )

            # run the example script in parallel
            run_parallel_script.run(
                filename=filename, args=parallel_args, nprocs=nprocs,
                timeout=timeout, path=MY_DIR
            )

            # get the serial and parallel results
            dir1path = os.path.abspath(dir1)
            dir2path = os.path.abspath(dir2)

            # load the serial output
            file = get_files(dirname=dir1path, fname=prefix)[-1]
            serial = load(file)
            serial = serial['arrays']['fluid']

            # load the parallel output
            file = get_files(dirname=dir2path, fname=prefix)[-1]
            parallel = load(file)
            parallel = parallel['arrays']['fluid']

        finally:
            shutil.rmtree(dir1, True)
            shutil.rmtree(dir2, True)

        # test
        self._test(serial, parallel, atol, nprocs)

    def _test(self, serial, parallel, atol, nprocs):
        # make sure the arrays are at the same time
        self.assertAlmostEqual(serial.time, parallel.time, 12)

        # test the results.
        xs, ys, zs, rhos = serial.get("x", "y", "z", "rho")
        xp, yp, zp, rhop, gid = parallel.get("x", "y", "z", "rho", 'gid')

        if nprocs == 1:
            # Not really a parallel run (used for openmp support).
            gid = np.arange(xs.size)

        self.assertTrue(xs.size, xp.size)

        x_err = np.max(xs[gid] - xp)
        y_err = np.max(ys[gid] - yp)
        z_err = np.max(zs[gid] - zp)
        self.assertTrue(np.allclose(xs[gid], xp, atol=atol, rtol=0),
                        "Max x_error %s" % x_err)
        self.assertTrue(np.allclose(ys[gid], yp, atol=atol, rtol=0),
                        "Max y_error %s" % y_err)
        self.assertTrue(np.allclose(zs[gid], zp, atol=atol, rtol=0),
                        "Max z_error %s" % z_err)



"""
Module contains some common functions.
"""

# standard imports
import errno
from glob import glob
import os
import socket
import sys
import time

import numpy

import pysph
from pysph.solver.output import load, dump, output_formats  # noqa: 401
from pysph.solver.output import gather_array_data as _gather_array_data

ASCII_FMT = " 123456789#"
try:
    uni_chr = unichr
except NameError:
    uni_chr = chr
UTF_FMT = u" " + u''.join(map(uni_chr, range(0x258F, 0x2587, -1)))


def _supports_unicode(fp):
    # Taken somewhat from the tqdm package.
    if not hasattr(fp, 'encoding'):
        return False
    else:
        encoding = fp.encoding
        try:
            u'\u2588\u2589'.encode(encoding)
        except UnicodeEncodeError:
            return False
        except Exception:
            try:
                return (encoding.lower().startswith('utf-')
                        or ('U8' == encoding))
            except:
                return False
        else:
            return True


def get_free_port(start, skip=None):
    """Return an integer that is an available port for a service. Start at the
    given `start` value and `skip` any specified values.
    """
    skip = () if skip is None else skip
    x = start
    while x < 65536:
        if x in skip:
            x += 1
        else:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                try:
                    s.bind(('', x))
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    return x
                except socket.error as e:
                    if e.errno == errno.EADDRINUSE:
                        x += 1
                    else:
                        raise


def is_using_ipython():
    """Return True if the code is being run from an IPython session or
    notebook.
    """
    try:
        # If this is being run inside an IPython console or notebook
        # then this is defined.
        __IPYTHON__
    except NameError:
        return False
    else:
        return True


def check_array(x, y):
    """Check if two arrays are equal with an absolute tolerance of
    1e-16."""
    return numpy.allclose(x, y, atol=1e-16, rtol=0)


def get_distributed_particles(pa, comm, cell_size):
    # FIXME: this can be removed once the examples all use Application.
    from pysph.parallel.load_balancer import LoadBalancer
    rank = comm.Get_rank()
    num_procs = comm.Get_size()

    if rank == 0:
        lb = LoadBalancer.distribute_particles(pa, num_procs=num_procs,
                                               block_size=cell_size)
    else:
        lb = None

    particles = comm.scatter(lb, root=0)

    return particles


def get_array_by_name(arrays, name):
    """Given a list of arrays and the name of the desired array, return the
    desired array.
    """
    for array in arrays:
        if array.name == name:
            return array


def fmt_time(time):
    mm, ss = divmod(time, 60)
    hh, mm = divmod(mm, 60)
    if hh > 0:
        s = "%d:%02d:%02d" % (hh, mm, ss)
    else:
        s = "%02d:%02.1f" % (mm, ss)
    return s


class ProgressBar(object):
    def __init__(self, ti, tf, show=True, file=None, ascii=False):
        if file is None:
            self.file = sys.stdout
        self.ti = ti
        self.tf = tf
        self.t = 0.0
        self.dt = 1.0
        self.start = time.time()
        self.count = 0
        self.iter_inc = 1
        self.show = show
        self.ascii = ascii
        if not ascii and not _supports_unicode(self.file):
            self.ascii = True
        if not self.file.isatty() and not is_using_ipython():
            self.show = False
        self.display()

    def _fmt_bar(self, percent, width):
        chars = ASCII_FMT if self.ascii else UTF_FMT
        nsyms = len(chars) - 1
        tens, ones = divmod(int(percent/100 * width * nsyms), nsyms)
        end = chars[ones] if ones > 0 else ''
        return (chars[-1]*tens + end).ljust(width)

    def _fmt_iters(self, iters):
        if iters < 1e3:
            s = '%d' % iters
        elif iters < 1e6:
            s = '%.1fk' % (iters/1e3)
        elif iters < 1e9:
            s = '%.1fM' % (iters/1e6)
        return s

    def display(self):
        if self.show:
            elapsed = time.time() - self.start
            if self.t > 0:
                eta = (self.tf - self.t)/self.t * elapsed
            else:
                eta = 0.0
            percent = int(round(self.t/self.tf*100))
            bar = self._fmt_bar(percent, 20)
            secsperit = elapsed/self.count if self.count > 0 else 0
            out = ('{percent:3}%|{bar}|'
                   ' {iters}it | {time:.1e}s [{elapsed}<{eta} | '
                   '{secsperit:.3f}s/it]').format(
                bar=bar, percent=percent, iters=self._fmt_iters(self.count),
                time=self.t, elapsed=fmt_time(elapsed), eta=fmt_time(eta),
                secsperit=secsperit
            )
            self.file.write('\r%s' % out.ljust(70))
            self.file.flush()

    def update(self, t, iter_inc=1):
        '''Set the current time and update the number of iterations.
        '''
        self.dt = t - self.t
        self.iter_inc = iter_inc
        self.count += iter_inc
        self.t = t
        self.display()

    def finish(self):
        self.display()
        if self.show:
            self.file.write('\n')


##############################################################################
# friendly mkdir  from http://code.activestate.com/recipes/82465/.
##############################################################################
def mkdir(newdir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    if os.path.isdir(newdir):
        pass

    elif os.path.isfile(newdir):
        raise OSError("a file with the same name as the desired "
                      "dir, '%s', already exists." % newdir)

    else:
        head, tail = os.path.split(newdir)

        if head and not os.path.isdir(head):
            mkdir(head)

        if tail:
            try:
                os.mkdir(newdir)
            # To prevent race in mpi runs
            except OSError as e:
                import errno
                if e.errno == errno.EEXIST and os.path.isdir(newdir):
                    pass
                else:
                    raise


def get_pysph_root():
    return os.path.split(pysph.__file__)[0]


def dump_v1(filename, particles, solver_data, detailed_output=False,
            only_real=True, mpi_comm=None):
    """Dump the given particles and solver data to the given filename using
    version 1.  This is mainly used only for testing that we can continue
    to load older versions of the data files.
    """

    all_array_data = {}
    output_data = {"arrays": all_array_data, "solver_data": solver_data}

    for array in particles:
        all_array_data[array.name] = array.get_property_arrays(
            all=detailed_output, only_real=only_real
        )

    # Gather particle data on root
    if mpi_comm is not None:
        all_array_data = _gather_array_data(all_array_data, mpi_comm)

    output_data['arrays'] = all_array_data

    if mpi_comm is None or mpi_comm.Get_rank() == 0:
        numpy.savez(filename, version=1, **output_data)


def load_and_concatenate(prefix, nprocs=1, directory=".", count=None):
    """Load the results from multiple files.

    Given a filename prefix and the number of processors, return a
    concatenated version of the dictionary returned via load.

    Parameters
    ----------

    prefix : str
        A filename prefix for the output file.

    nprocs : int
        The number of processors (files) to read

    directory : str
        The directory for the files

    count : int
        The file iteration count to read. If None, the last available
        one is read

    """

    if count is None:
        counts = [i.rsplit('_', 1)[1][:-4] for i in os.listdir(directory)
                  if i.startswith(prefix) and i.endswith('.npz')]
        counts = sorted([int(i) for i in counts])
        count = counts[-1]

    arrays_by_rank = {}

    for rank in range(nprocs):
        fname = os.path.join(
            directory, prefix + '_' + str(rank) + '_' + str(count) + '.npz'
        )

        data = load(fname)
        arrays_by_rank[rank] = data["arrays"]

    arrays = _concatenate_arrays(arrays_by_rank, nprocs)

    data["arrays"] = arrays

    return data


def _concatenate_arrays(arrays_by_rank, nprocs):
    """Concatenate arrays into one single particle array. """

    if nprocs <= 0:
        return 0

    array_names = arrays_by_rank[0].keys()
    first_processors_arrays = arrays_by_rank[0]

    if nprocs > 1:
        ret = {}
        for array_name in array_names:
            first_array = first_processors_arrays[array_name]
            for rank in range(1, nprocs):
                other_processors_arrays = arrays_by_rank[rank]
                other_array = other_processors_arrays[array_name]

                # append the other array to the first array
                first_array.append_parray(other_array)

                # remove the non local particles
                first_array.remove_tagged_particles(1)

            ret[array_name] = first_array

    else:
        ret = arrays_by_rank[0]

    return ret


def get_files(dirname=None, fname=None, endswith=output_formats):
    """Get all solution files in a given directory, `dirname`.

    Parameters
    ----------

    dirname: str
        Name of directory.
    fname: str
        An initial part of the filename, if not specified use the first
        part of the dirname.
    endswith: str
        The extension of the file to load.
    """

    if dirname is None:
        return []

    path = os.path.abspath(dirname)

    if fname is None:
        infos = glob(os.path.join(path, "*.info"))
        if infos:
            fname = os.path.splitext(os.path.basename(infos[0]))[0]
        else:
            fname = os.path.basename(path).split('_output')[0]

    files = glob(os.path.join(path, "%s*.*" % fname))
    files = [f for f in files if f.endswith(endswith)]

    # sort the files
    files.sort(key=_sort_key)

    return files


def iter_output(files, *arrays):
    """Given an iterable of the solution files, this loads the files, and
    yields the solver data and the requested arrays.

    If arrays is not supplied, it returns a dictionary of the arrays.

    Parameters
    ----------

    files : iterable
        Iterates over the list of desired files

    *arrays : strings
        Optional series of array names of arrays to return.

    Examples
    --------

    >>> files = get_files('elliptical_drop_output')
    >>> for solver_data, arrays in iter_output(files):
    ...     print(solver_data['t'], arrays.keys())

    >>> files = get_files('elliptical_drop_output')
    >>> for solver_data, fluid in iter_output(files, 'fluid'):
    ...     print(solver_data['t'], fluid.name)

    """
    for file in files:
        data = load(file)
        solver_data = data['solver_data']
        if len(arrays) == 0:
            yield solver_data, data['arrays']
        else:
            _arrays = [data['arrays'][x] for x in arrays]
            yield [solver_data] + _arrays


def _sort_key(arg):
    a = os.path.splitext(arg)[0]
    return int(a[a.rfind('_') + 1:])


def remove_irrelevant_files(files):
    """Remove any npz files that are not output files.

    That is, the file should not end with a '_number.npz'. This allows users to
    dump other .npz of .hdf5 files in the output while post-processing without
    breaking.
    """
    result = []
    for f in files:
        try:
            _sort_key(f)
        except ValueError:
            pass
        else:
            result.append(f)
    return result



"""Two-dimensional dam break over a dry bed.  (30 minutes)

The case is described in "B.Buchner, Green Water on Ship -
Type Off shore Structures, Ph.D.thesis,TU Delft,
Delft University of Technology (2002), Appendix II"

URL: https://repository.tudelft.nl/islandora/object/uuid%3Af0c0bd67-d52a-4b79-8451-1279629a5b80

"""

import os
import numpy as np

from pysph.base.utils import get_particle_array
from pysph.tools.geometry import get_2d_tank, get_2d_block
from pysph.examples.dam_break_2d import DamBreak2D

fluid_column_height = 1.0
H = fluid_column_height
fluid_column_width = 2.0 * H
container_height = 3.0 * H
container_width = 5.366 * H
nboundary_layers = 2
nu = 0.0
dx = 0.03
g = 9.81
ro = 1000.0
vref = np.sqrt(2 * 9.81 * H)
co = 10.0 * vref
gamma = 7.0
alpha = 0.1
beta = 0.0
B = co * co * ro / gamma
p0 = 1000.0
hdx = 1.3
h = hdx * dx
m = dx**2 * ro


class DamBreak2DBuchner(DamBreak2D):
    def add_user_options(self, group):
        super(DamBreak2DBuchner, self).add_user_options(group)

        interp_methods = ['shepard', 'sph', 'order1']
        group.add_argument(
            '--interp-method', action="store", type=str, dest='interp_method',
            default='shepard', help="Specify the interpolation method.",
            choices=interp_methods
        )

    def consume_user_options(self):
        super(DamBreak2DBuchner, self).consume_user_options()
        self.interp_method = self.options.interp_method
        if self.options.scheme != "iisph":
            self.co = co
            self.scheme.configure(c0=co)

    def create_particles(self):
        if self.options.staggered_grid:
            nboundary_layers = 2
            nfluid_offset = 2
            wall_hex_pack = True
        else:
            nboundary_layers = 4
            nfluid_offset = 1
            wall_hex_pack = False

        xt, yt = get_2d_tank(
                             dx=self.dx,
                             length=container_width,
                             height=container_height,
                             base_center=[container_width/2, 0],
                             num_layers=nboundary_layers
                             )

        xf, yf = get_2d_block(
                              dx=self.dx,
                              length=fluid_column_width,
                              height=fluid_column_height,
                              center=[fluid_column_width/2, fluid_column_height/2]
                              )

        xf += self.dx
        yf += self.dx

        fluid = get_particle_array(name='fluid', x=xf, y=yf, h=h, m=m, rho=ro)
        boundary = get_particle_array(name='boundary', x=xt, y=yt, h=h, m=m,
                                      rho=ro)

        self.scheme.setup_properties([fluid, boundary])
        if self.options.scheme == 'iisph':
            # the default position tends to cause the particles to be pushed
            # away from the wall, so displacing it by a tiny amount helps.
            fluid.x += self.dx / 4

        self.kernel_corrections(fluid, boundary)

        return [fluid, boundary]

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        import matplotlib.pyplot as plt
        from pysph.examples import db_exp_data as dbd
        from pysph.tools.interpolator import Interpolator
        H = 1.0
        factor_y = 1/(ro*g*H)
        factor_x = np.sqrt(g/H)

        data_t, data_p0 = dbd.get_buchner_data()
        files = self.output_files

        t = []
        p0 = []
        for sd, arrays1, arrays2 in iter_output(files, "fluid", "boundary"):
            t.append(sd["t"]*factor_x)
            interp = Interpolator([arrays1, arrays2], x=[container_width],
                                  y=[H*0.2], method=self.interp_method)
            p0.append(interp.interpolate('p')*factor_y)

        plt.plot(t, p0, label="Computed")

        fname = os.path.join(self.output_dir, 'results.npz')
        t, p0 = list(map(np.asarray, (t, p0)))
        np.savez(fname, t=t, p0=p0)

        plt.scatter(data_t, data_p0, color=(0, 0, 0), label="Experiment (Buchner, 2002)")
        plt.legend()
        plt.ylabel(r"$\frac{P}{\rho gH}$")
        plt.xlabel(r"$t \sqrt{\frac{g}{H}}$")
        plt.savefig(os.path.join(self.output_dir, 'p_vs_t.png'))


if __name__ == '__main__':
    app = DamBreak2DBuchner()
    app.run()
    app.post_process(app.info_filename)



import numpy as np
import pyopencl as cl
import pyopencl.array
import pyopencl.cltypes
from pyopencl.elementwise import ElementwiseKernel
from pytools import memoize
from compyle.opencl import get_context
from pysph.base.gpu_nnps_helper import GPUNNPSHelper
from compyle.array import Array

make_vec_dict = {
    'float': {
        1: np.float32,
        2: cl.cltypes.make_float2,
        3: cl.cltypes.make_float3
    },
    'double': {
        1: np.float64,
        2: cl.cltypes.make_double2,
        3: cl.cltypes.make_double3
    }
}


@memoize
def get_helper(src_file, c_type=None):
    # ctx and c_type are the only parameters that
    # change here
    return GPUNNPSHelper(src_file, backend='opencl',
                         c_type=c_type)


@memoize
def get_copy_kernel(ctx, dtype1, dtype2, varnames):
    arg_list = [('%(data_t1)s *%(v)s1' % dict(data_t1=dtype1, v=v))
                for v in varnames]
    arg_list += [('%(data_t2)s *%(v)s2' % dict(data_t2=dtype2, v=v))
                 for v in varnames]
    args = ', '.join(arg_list)

    operation = '; '.join(('%(v)s2[i] = (%(data_t2)s)%(v)s1[i];' %
                           dict(v=v, data_t2=dtype2))
                          for v in varnames)
    return ElementwiseKernel(ctx, args, operation=operation)


_vector_dtypes = {
    'uint': {
        2: cl.cltypes.uint2,
        4: cl.cltypes.uint4,
        8: cl.cltypes.uint8
    },
    'float': {
        1: cl.cltypes.float,
        2: cl.cltypes.float2,
        3: cl.cltypes.float3,
    },
    'double': {
        1: cl.cltypes.double,
        2: cl.cltypes.double2,
        3: cl.cltypes.double3
    }
}


def get_vector_dtype(ctype, dim):
    try:
        return _vector_dtypes[ctype][dim]
    except KeyError:
        raise ValueError("Vector datatype of type %(ctype)s with %(dim)s items"
                         " is not supported" % dict(ctype=ctype, dim=dim))


c2d = {
    'half': np.float16,
    'float': np.float32,
    'double': np.float64
}


def ctype_to_dtype(ctype):
    return c2d[ctype]


class GPUParticleArrayWrapper(object):
    def __init__(self, pa_gpu, c_type_src, c_type, varnames):
        self.c_type = c_type
        self.c_type_src = c_type_src
        self.varnames = varnames
        self._allocate_memory(pa_gpu)
        self.sync(pa_gpu)

    def _gpu_copy(self, pa_gpu):
        copy_kernel = get_copy_kernel(get_context(), self.c_type_src,
                                      self.c_type, self.varnames)
        args = [getattr(pa_gpu, v).dev for v in self.varnames]
        args += [getattr(self, v).dev for v in self.varnames]
        copy_kernel(*args)

    def _allocate_memory(self, pa_gpu):
        shape = getattr(pa_gpu, self.varnames[0]).dev.shape[0]
        for v in self.varnames:
            setattr(self, v,
                    Array(ctype_to_dtype(self.c_type),
                          n=shape, backend='opencl'))

    def _gpu_sync(self, pa_gpu):
        v0 = self.varnames[0]

        if getattr(self, v0).dev.shape != getattr(pa_gpu, v0).dev.shape:
            self._allocate_memory(pa_gpu)
        self._gpu_copy(pa_gpu)

    def sync(self, pa_gpu):
        self._gpu_sync(pa_gpu)


class ParticleArrayWrapper(object):
    """A loose wrapper over Particle Array

    Objective is to transparently maintain a copy of
    the original particle array's position properties
    (x, y, z, h)
    """

    def __init__(self, pa, c_type_src, c_type, varnames):
        self._pa = pa
        # If data types are different, then make a copy of the
        # underlying data stored on the device
        if c_type_src != c_type:
            self._pa_gpu_is_copy = True
            self._gpu = GPUParticleArrayWrapper(pa.gpu, c_type_src,
                                                c_type, varnames)
        else:
            self._pa_gpu_is_copy = False
            self._gpu = None

    def get_number_of_particles(self):
        return self._pa.get_number_of_particles()

    @property
    def gpu(self):
        if self._pa_gpu_is_copy:
            return self._gpu
        else:
            return self._pa.gpu

    def sync(self):
        if self._pa_gpu_is_copy:
            self._gpu.sync(self._pa.gpu)



from unittest import TestCase, main

from ..utils import is_overloaded_method


class TestUtils(TestCase):
    def test_is_overloaded_method_works_for_simple_overloads(self):
        # Given
        class A(object):
            def f(self): pass

        class B(A):
            pass

        # When/Then
        b = B()
        self.assertFalse(is_overloaded_method(b.f))

        class C(A):
            def f(self): pass

        # When/Then
        c = C()
        self.assertTrue(is_overloaded_method(c.f))

    def test_is_overloaded_method_works_for_parent_overloads(self):
        # Given
        class A(object):
            def f(self): pass

        class B(A):
            def f(self): pass

        class C(B):
            pass

        # When/Then
        c = C()
        self.assertTrue(is_overloaded_method(c.f))


if __name__ == '__main__':
    main()



from pysph.base.particle_array import ParticleArray
from pysph.base import nnps
from pysph.tools.geometry import remove_repeated_points
import numpy as np
from numpy.linalg import norm
from cyarray.api import UIntArray
cimport numpy as np

DTYPE = np.float64
ctypedef np.float_t DTYPE_t


class ZeroAreaTriangleException(Exception):
    pass


class PolygonMeshError(ValueError):
    pass


cpdef _point_sign(double x1, double y1, double x2,
                double y2, double x3, double y3):
    return (x1 - x3) * (y2 - y3) - (x2 - x3) * (y1 - y3)


def _in_triangle(px, py, vx1, vy1, vx2, vy2, vx3, vy3):
    b1 = _point_sign(px, py, vx1, vy1, vx2, vy2) < 0
    b2 = _point_sign(px, py, vx2, vy2, vx3, vy3) < 0
    b3 = _point_sign(px, py, vx3, vy3, vx1, vy1) < 0

    return ((b1 == b2) and (b2 == b3))


def _interp_2d(v0, v1, dx):
    l = norm(v0 - v1)
    p = np.linspace(0., 1., int(l / dx + 2)).reshape(-1, 1)
    return v0.reshape(1, -1) * p + v1.reshape(1, -1) * (1. - p)


def _get_triangle_sides(triangle, dx):
    """Interpolate points on the sides of the triangle"""
    sides = np.vstack([_interp_2d(triangle[0], triangle[1], dx),
                       _interp_2d(triangle[1], triangle[2], dx),
                       _interp_2d(triangle[2], triangle[0], dx)])
    return sides[:, 0], sides[:, 1], sides[:, 2]


def _projection_parameters(x, a, b, p):
    """Parameters of projection of x on surface (sa + tb + p)
    """
    return np.dot(x - p, a), np.dot(x - p, b)


def _fill_triangle(triangle, h=0.1):
    EPS = np.finfo(float).eps
    if triangle.shape[0] != 3:
        raise PolygonMeshError("non-triangular meshes are not supported: "
                               "dim({}) ! = 3.".format(triangle))

    if norm(triangle[0] - triangle[1]) < EPS or \
       norm(triangle[1] - triangle[2]) < EPS or \
       norm(triangle[2] - triangle[0]) < EPS:
        raise ZeroAreaTriangleException(
            "unable to interpolate in zero area triangle: {}".format(triangle))

    # Surface will be of the form v = sa + tb + p
    # p is (arbitrarily) taken as the centroid of the triangle

    p = (triangle[0] + triangle[1] + triangle[2]) / 3
    a = (triangle[1] - triangle[0]) / norm(triangle[1] - triangle[0])
    b = np.cross(np.cross(a, triangle[2] - triangle[1]), a)
    b /= norm(b)
    st = np.array([_projection_parameters(triangle[0], a, b, p),
                   _projection_parameters(triangle[1], a, b, p),
                   _projection_parameters(triangle[2], a, b, p)])

    st_min, st_max = np.min(st, axis=0), np.max(st, axis=0)
    s_mesh, t_mesh = np.meshgrid(np.arange(st_min[0], st_max[0], h),
                                 np.arange(st_min[1], st_max[1], h))

    s_mesh, t_mesh = s_mesh.ravel(), t_mesh.ravel()
    mask = np.empty(len(s_mesh), dtype='bool')

    for i in range(len(s_mesh)):
        mask[i] = _in_triangle(s_mesh[i], t_mesh[i],
                               st[0, 0], st[0, 1],
                               st[1, 0], st[1, 1],
                               st[2, 0], st[2, 1])

    s_mesh, t_mesh = s_mesh[mask], t_mesh[mask]

    # Final mesh coordinates generated from parameters s and t

    result = np.dot(s_mesh.reshape(-1, 1), a.reshape(-1, 1).T) + \
        np.dot(t_mesh.reshape(-1, 1), b.reshape(-1, 1).T) + p
    return result[:, 0], result[:, 1], result[:, 2]


def _get_surface_mesh(x, y, z, cells, dx_triangle, uniform = False):
    """Interpolates points within triangles spaced by dx_triangle"""
    vectors = np.zeros((len(cells), 3, 3))
    x_list, y_list, z_list = [], [], []
    for i, cell in enumerate(cells):
        idx1, idx2, idx3 = cell

        vector = np.array([[x[idx1], y[idx1], z[idx1]],
                           [x[idx2], y[idx2], z[idx2]],
                           [x[idx3], y[idx3], z[idx3]]])

        x1, y1, z1 = _fill_triangle(vector, dx_triangle)
        x2, y2, z2 = _get_triangle_sides(vector, dx_triangle)
        x_list.append(np.r_[x1, x2])
        y_list.append(np.r_[y1, y2])
        z_list.append(np.r_[z1, z2])

        vectors[i] = vector

    xm = np.concatenate(x_list)
    ym = np.concatenate(y_list)
    zm = np.concatenate(z_list)

    if uniform:
        return xm, ym, zm, x_list, y_list, z_list, vectors
    else:
        return xm, ym, zm


def prism(tri_normal, tri_points, dx_sph):
    """
    Parameters
    ----------
    tri_normal : outward normal of triangle
    tri_points : points forming the triangle
    dx_sph :  grid spacing
    Returns
    -------
    prism_normals : 5X3 array containing the 5 outward normals of the prism
    prism _points : 6X3 array containing the 6 points used to define the prism
    prism_face_centres : 5X3 array containing the centres of the 5 faces
                         of the prism
    """
    cdef np.ndarray prism_normals = np.zeros((5, 3), dtype=DTYPE)
    cdef np.ndarray prism_points = np.zeros((6, 3), dtype=DTYPE)
    cdef np.ndarray prism_face_centres = np.zeros((5, 3), dtype=DTYPE)
    cdef int sign = 1
    cdef int m, n

    # unit normals of the triangular faces of the prism.
    prism_normals[0] = tri_normal / norm(tri_normal)
    prism_normals[4] = -1 * prism_normals[0]

    # distance between triangular faces of prism
    cdef float h = 1.5 * dx_sph

    for m in range(3):
        prism_points[m] = tri_points[m]

    # second triangular face at a distance h from triangle.
    for m in range(3):
        for n in range(3):
            prism_points[m+3][n] = tri_points[m][n] - tri_normal[n]*h

    #need to determine if point orientation is clockwise or anticlockwise
    #to determine normals direction.
    normal_tri_cross = np.cross(prism_points[2]-prism_points[0],
                                prism_points[1]-prism_points[0])
    if np.dot(tri_normal, normal_tri_cross) < 0:
        sign = 1  # clockwise
    else:
        sign = -1  # anti-clockwise

    # Normals of the reactangular faces of the prism.
    prism_normals[1] = sign * np.cross(prism_points[1]-prism_points[0],
                                       prism_points[1]-prism_points[4])

    prism_normals[2] = sign * np.cross(prism_points[2]-prism_points[0],
                                       prism_points[5]-prism_points[2])

    prism_normals[3] = sign * np.cross(prism_points[1]-prism_points[2],
                                       prism_points[5]-prism_points[2])

    # centroids of the triangles
    prism_face_centres[0] = (prism_points[0] +
                             prism_points[1] +
                             prism_points[2])/3

    prism_face_centres[4] = (prism_points[3] +
                             prism_points[4] +
                             prism_points[5])/3

    # centres of the rectangular faces
    prism_face_centres[1] = (prism_points[0] + prism_points[3] +
                             prism_points[1] + prism_points[4])/4

    prism_face_centres[2] = (prism_points[0] + prism_points[3] +
                             prism_points[2] + prism_points[5])/4

    prism_face_centres[3] = (prism_points[1] + prism_points[2] +
                             prism_points[4] + prism_points[5])/4

    return prism_normals, prism_points, prism_face_centres


def get_points_from_mgrid(pa_grid, pa_mesh, x_list, y_list, z_list,
                          float radius_scale, float dx_sph, vectors,
                          normals):
    """
    The function finds nearest neighbours for a given mesh on a given grid
    and only returns those points which lie within the volume of the
    object
    Parameters
    ----------
    pa_grid : Source particle array
    pa_mesh : Destination particle array
    x_list, y_list, z_list : Coordinates of surface points for each triangle
    """
    pa_list = [pa_mesh, pa_grid]
    nps = nnps.LinkedListNNPS(dim=3, particles=pa_list,
                              radius_scale=radius_scale)
    cdef int src_index = 1, dst_index = 0
    nps.set_context(src_index=1, dst_index=0)
    nbrs = UIntArray()
    cdef np.ndarray prism_normals = np.zeros((5, 3), dtype=DTYPE)
    cdef np.ndarray prism_face_centres = np.zeros((5, 3), dtype=DTYPE)
    cdef np.ndarray prism_points = np.zeros((6, 3), dtype=DTYPE)
    cdef list selected_points = []
    cdef int counter = 0, l = 0
    # Iterating over each triangle
    for i in range(np.shape(x_list)[0]):
        prism_normals, prism_points, prism_face_centres = prism(
            normals[i], vectors[i], dx_sph
        )
        # Iterating over surface points in triangle to find nearest
        # neighbour on grid.
        for j in range(len(x_list[i])):
            nps.get_nearest_particles(src_index, dst_index, counter, nbrs)
            neighbours = nbrs.get_npy_array()
            l = len(neighbours)
            for t in range(l):
                point = np.array([pa_grid.x[neighbours[t]],
                                  pa_grid.y[neighbours[t]],
                                  pa_grid.z[neighbours[t]]])
                # determining whether point is within prism.
                if inside_prism(point, prism_normals,
                                prism_points, prism_face_centres):
                    selected_points.append(neighbours[t])
            counter = counter + 1
    idx_set = set(selected_points)
    idx = list(idx_set)
    return pa_grid.x[idx], pa_grid.y[idx], pa_grid.z[idx]


cdef bint inside_prism(double[:] point, double[:,:] prism_normals,
                        double[:,:] prism_points, double[:,:] prism_face_centres):
    """ Identifies whether a point is within the corresponding prism by checking
        if all dot products of the normals of the prism with the vector joining
        the point and a point on the corresponding side is negative
    """
    if dot(prism_normals[0], point, prism_face_centres[0]) > 0:
        return False
    if dot(prism_normals[4], point, prism_face_centres[4]) > 0:
        return False
    if dot(prism_normals[1], point, prism_face_centres[1]) > 0:
        return False
    if dot(prism_normals[2], point, prism_face_centres[2]) > 0:
        return False
    if dot(prism_normals[3], point, prism_face_centres[3]) > 0:
        return False
    return True


cdef double dot(double[:] normal, double[:] point, double[:] face_centre):
    return normal[0]*(point[0]-face_centre[0]) + \
           normal[1]*(point[1]-face_centre[1]) + \
           normal[2]*(point[2]-face_centre[2])


def surf_points_uniform(x, y, z, cells, normals, dx_sph=1,
                        h_sph=1, radius_scale=1.0,
                        dx_triangle=None):
    """Generates points to cover surface described by a set of points
    The function generates a grid with a spacing of dx_sph and keeps points
    on the grid which lie within the object.

    The algorithm for this is consists of the following
    steps:
    1. Interpolate a set of points over the triangles
    2. Create a grid that covers the entire object
    3. Remove grid points generated outside the given object by checking
       if the points lie within prisms formed by the triangles.
    Parameters
    ----------
    x : ndarray
        1d numpy array with x coordinates of surface points
    y : ndarray
        1d numpy array with y coordinates of surface points
    z : ndarray
        1d numpy array with z coordinates of surface points
    cells : ndarray
        2d numpy array of triplets containing vertices of the triangles as indices
    normals : ndarray
        2d numpy array of triplets containing outward normals at each cell
    dx_sph : float
        Spacing in generated grid points
    h_sph : float
        Smoothing length
    radius_scale : float
        Kernel radius scale
    dx_triangle : float, optional
        By default, dx_triangle = 0.5 * dx_sph
    Returns
    -------
    x : ndarray
        1d numpy array with x coordinates of surface grid
    y : ndarray
        1d numpy array with y coordinates of surface grid
    z : ndarray
        1d numpy array with z coordinates of surface grid
    Raises
    ------
    PolygonMeshError
        If polygons in file are not all triangles
    """
    if dx_triangle is None:
        dx_triangle = 0.5 * dx_sph

    xm, ym, zm, x_list, y_list, z_list, vectors = \
        _get_surface_mesh(x, y, z, cells, dx_triangle, uniform = True)

    pa_mesh = ParticleArray(name='mesh', x=xm, y=ym, z=zm, h=h_sph)
    offset = radius_scale * h_sph

    x_grid, y_grid, z_grid = np.meshgrid(
        np.arange(xm.min() - offset, xm.max() + offset, dx_sph),
        np.arange(ym.min() - offset, ym.max() + offset, dx_sph),
        np.arange(zm.min() - offset, zm.max() + offset, dx_sph)
    )

    pa_grid = ParticleArray(name='grid', x=x_grid, y=y_grid, z=z_grid, h=h_sph)
    xf, yf, zf = get_points_from_mgrid(pa_grid, pa_mesh, x_list,
                                       y_list, z_list, radius_scale,
                                       dx_sph, vectors, normals)
    return xf, yf, zf


def surface_points(x, y, z, cells, dx_triangle):
    """ Generates points to cover surface described by given set of connected points
    Parameters
    ----------
    x : ndarray
        1d numpy array with x coordinates of surface points
    y : ndarray
        1d numpy array with y coordinates of surface points
    z : ndarray
        1d numpy array with z coordinates of surface points
    cells : ndarray
        2d numpy array of triplets containing vertices of the triangles as indices
    dx_triangle : float
        Required spacing between generated particles
    Returns
    -------
    xf : ndarray
        1d numpy array with x coordinates of covered surface
    yf : ndarray
        1d numpy array with y coordinates of covered surface
    zf : ndarray
        1d numpy array with z coordinates of covered surface
    """

    xm, ym, zm = _get_surface_mesh(x, y, z, cells, dx_triangle, uniform = False)
    xf, yf, zf = remove_repeated_points(xm, ym, zm, dx_triangle)
    return xf, yf, zf



# cython: language_level=3, embedsignature=True
# distutils: language=c++
# Library imports.
import numpy as np
cimport numpy as np

# Cython imports
from cython.operator cimport dereference as deref, preincrement as inc
from cython.parallel import parallel, prange, threadid

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.vector cimport vector

# cpython
from cpython.dict cimport PyDict_Clear, PyDict_Contains, PyDict_GetItem
from cpython.list cimport PyList_GetItem, PyList_SetItem, PyList_GET_ITEM

# Cython for compiler directives
cimport cython


#############################################################################
cdef class LinkedListNNPS(NNPS):
    """Nearest neighbor query class using the linked list method.
    """
    def __init__(self, int dim, list particles, double radius_scale=2.0,
                 int ghost_layers=1, domain=None,
                 bint fixed_h=False, bint cache=False, bint sort_gids=False):
        """Constructor for NNPS

        Parameters
        ----------

        dim : int
            Number of dimension.

        particles : list
            The list of particle arrays we are working on

        radius_scale : double, default (2)
            Optional kernel radius scale. Defaults to 2

        ghost_layers : int
            Optional number of layers to share in parallel

        domain : DomainManager, default (None)
            Optional limits for the domain

        fixed_h : bint
            Optional flag to use constant cell sizes throughout.

        cache : bint
            Flag to set if we want to cache neighbor calls. This costs
            storage but speeds up neighbor calculations.

        sort_gids : bint, default (False)
            Flag to sort neighbors using gids (if they are available).
            This is useful when comparing parallel results with those
            from a serial run.
        """
        # initialize the base class
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        # initialize the head and next for each particle array
        self.heads = [UIntArray() for i in range(self.narrays)]
        self.nexts = [UIntArray() for i in range(self.narrays)]

        # flag for constant smoothing lengths
        self.fixed_h = fixed_h

        # defaults
        self.ncells_per_dim = IntArray(3)
        self.n_cells = 0
        self.sort_gids = sort_gids

        # compute the intial box sort for all local particles. The
        # DomainManager.setup_domain method is called to compute the
        # cell size.
        self.domain.update()
        self.update()

    #### Public protocol ################################################

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        # Number of cells
        cdef int n_cells = self.n_cells
        cdef int dim = self.dim

        # cell shifts
        cdef int* shifts = self.cell_shifts.data

        # Source data arrays
        cdef double* s_x = self.src.x.data
        cdef double* s_y = self.src.y.data
        cdef double* s_z = self.src.z.data
        cdef double* s_h = self.src.h.data
        cdef unsigned int* s_gid = self.src.gid.data

        # Destination particle arrays
        cdef double* d_x = self.dst.x.data
        cdef double* d_y = self.dst.y.data
        cdef double* d_z = self.dst.z.data
        cdef double* d_h = self.dst.h.data
        cdef unsigned int* d_gid = self.dst.gid.data

        cdef unsigned int* head = self.head.data
        cdef unsigned int* next = self.next.data

        # minimum values for the particle distribution
        cdef double* xmin = self.xmin.data

        # cell size and radius
        cdef double radius_scale = self.radius_scale
        cdef double cell_size = self.cell_size

        # locals
        cdef size_t indexj
        cdef double xij2
        cdef double hi2, hj2
        cdef int ierr, nnbrs
        cdef unsigned int _next
        cdef int ix, iy, iz

        # this is the physical position of the particle that will be
        # used in pairwise searching
        cdef double x = d_x[d_idx]
        cdef double y = d_y[d_idx]
        cdef double z = d_z[d_idx]

        # get the un-flattened index for the destination particle with
        # respect to the minimum
        cdef int _cid_x, _cid_y, _cid_z
        find_cell_id_raw(
            x - xmin[0], y - xmin[1], z - xmin[2],
            cell_size, &_cid_x, &_cid_y, &_cid_z
        )

        cdef int cid_x, cid_y, cid_z
        cdef long cell_index, orig_length
        cid_x = cid_y = cid_z = 0

        # gather search radius
        hi2 = radius_scale * d_h[d_idx]
        hi2 *= hi2

        orig_length = nbrs.length

        # Begin search through neighboring cells
        for ix in range(3):
            for iy in range(3):
                for iz in range(3):
                    cid_x = _cid_x + shifts[ix]
                    cid_y = _cid_y + shifts[iy]
                    cid_z = _cid_z + shifts[iz]

                    # Only consider valid cell indices
                    cell_index = self._get_valid_cell_index(
                        cid_x, cid_y, cid_z,
                        self.ncells_per_dim.data, dim, n_cells
                    )
                    if cell_index > -1:

                        # get the first particle and begin iteration
                        _next = head[ cell_index ]
                        while( _next != UINT_MAX ):
                            hj2 = radius_scale * s_h[_next]
                            hj2 *= hj2

                            xij2 = norm2( s_x[_next]-x,
                                          s_y[_next]-y,
                                          s_z[_next]-z )

                            # select neighbor
                            if ( (xij2 < hi2) or (xij2 < hj2) ):
                                nbrs.c_append(_next)

                            # get the 'next' particle in this cell
                            _next = next[_next]
        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices):
        cdef UIntArray head = self.heads[pa_index]
        cdef UIntArray next = self.nexts[pa_index]
        cdef ZOLTAN_ID_TYPE _next
        indices.reset()
        cdef long i

        for i in range(self.n_cells):
            _next = head.data[i]
            while (_next != UINT_MAX):
                indices.append(<long>_next)
                _next = next.data[_next]

    cpdef set_context(self, int src_index, int dst_index):
        """Setup the context before asking for neighbors.  The `dst_index`
        represents the particles for whom the neighbors are to be determined
        from the particle array with index `src_index`.

        Parameters
        ----------

         src_index: int: the source index of the particle array.
         dst_index: int: the destination index of the particle array.
        """
        NNPS.set_context(self, src_index, dst_index)

        # Set the current context.
        self.src = self.pa_wrappers[ src_index ]
        self.dst = self.pa_wrappers[ dst_index ]

        # next and head linked lists
        self.next = self.nexts[ src_index ]
        self.head = self.heads[ src_index ]


    #### Private protocol ################################################

    cpdef _bin(self, int pa_index, UIntArray indices):
        """Bin a given particle array with indices.

        Parameters
        ----------

        pa_index : int
            Index of the particle array corresponding to the particles list

        indices : UIntArray
            Subset of particles to bin

        """
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[ pa_index ]
        cdef DoubleArray x = pa_wrapper.x
        cdef DoubleArray y = pa_wrapper.y
        cdef DoubleArray z = pa_wrapper.z

        cdef DoubleArray xmin = self.xmin
        cdef DoubleArray xmax = self.xmax
        cdef IntArray ncells_per_dim = self.ncells_per_dim
        cdef int dim = self.dim

        # the head and next arrays for this particle array
        cdef UIntArray head = self.heads[ pa_index ]
        cdef UIntArray next = self.nexts[ pa_index ]
        cdef double cell_size = self.cell_size

        cdef UIntArray lindices, gindices
        cdef size_t num_particles, indexi, i

        # point and flattened index
        cdef cPoint pnt = cPoint_new(0, 0, 0)
        cdef int _cid

        # now bin the particles
        num_particles = indices.length
        for indexi in range(num_particles):
            i = indices.data[indexi]

            # the flattened index is considered relative to the
            # minimum along each co-ordinate direction
            pnt.x = x.data[i] - xmin.data[0]
            pnt.y = y.data[i] - xmin.data[1]
            pnt.z = z.data[i] - xmin.data[2]

            # flattened cell index
            _cid = self._get_flattened_cell_index(pnt, cell_size)

            # insert this particle
            next.data[ i ] = head.data[ _cid ]
            head.data[_cid] = i

    cdef long _get_flattened_cell_index(self, cPoint pnt, double cell_size):
        return flatten(
            find_cell_id(pnt, cell_size), self.ncells_per_dim, self.dim
        )

    cpdef long _get_number_of_cells(self) except -1:
        cdef double cell_size = self.cell_size
        cdef double cell_size1 = 1./cell_size
        cdef DoubleArray xmin = self.xmin
        cdef DoubleArray xmax = self.xmax
        cdef int ncx, ncy, ncz
        cdef long _ncells
        cdef int dim = self.dim

        # calculate the number of cells.
        ncx = <int>ceil( cell_size1*(xmax.data[0] - xmin.data[0]) )
        ncy = <int>ceil( cell_size1*(xmax.data[1] - xmin.data[1]) )
        ncz = <int>ceil( cell_size1*(xmax.data[2] - xmin.data[2]) )

        if ncx < 0 or ncy < 0 or ncz < 0:
            msg = 'LinkedListNNPS: Number of cells is negative '\
                   '(%s, %s, %s).'%(ncx, ncy, ncz)
            raise RuntimeError(msg)
            return -1

        ncx = 1 if ncx == 0 else ncx
        ncy = 1 if ncy == 0 else ncy
        ncz = 1 if ncz == 0 else ncz

        # number of cells along each coordinate direction
        self.ncells_per_dim.data[0] = ncx
        self.ncells_per_dim.data[1] = ncy
        self.ncells_per_dim.data[2] = ncz

        # total number of cells
        _ncells = ncx
        if dim == 2: _ncells = ncx * ncy
        if dim == 3: _ncells = ncx * ncy * ncz
        return _ncells

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef long _get_valid_cell_index(self, int cid_x, int cid_y, int cid_z,
            int* ncells_per_dim, int dim, int n_cells) nogil:
        return get_valid_cell_index(
            cid_x, cid_y, cid_z, ncells_per_dim, dim, n_cells
        )

    cpdef long _count_occupied_cells(self, long n_cells) except -1:
        if n_cells < 0 or n_cells > 2**28:
            # unsigned ints are 4 bytes, which means 2**28 cells requires 1GB.
            msg = "ERROR: LinkedListNNPS requires too many cells (%s)."%n_cells
            raise RuntimeError(msg)
            return -1

        return n_cells

    cpdef _refresh(self):
        """Refresh the head and next arrays locally"""
        cdef DomainManager domain = self.domain
        cdef int narrays = self.narrays
        cdef list pa_wrappers = self.pa_wrappers
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef list heads = self.heads
        cdef list nexts = self.nexts

        # locals
        cdef int i, j, np
        cdef long _ncells
        cdef UIntArray head, next

        _ncells = self._get_number_of_cells()

        _ncells = self._count_occupied_cells(_ncells)

        self.n_cells = <int>_ncells

        # initialize the head and next arrays
        for i in range(narrays):
            pa_wrapper = pa_wrappers[i]
            np = pa_wrapper.get_number_of_particles()

            # re-size the head and next arrays
            head = <UIntArray>PyList_GetItem(heads, i)
            next = <UIntArray>PyList_GetItem(nexts, i )

            head.resize( _ncells )
            next.resize( np )

            # UINT_MAX is used to indicate an invalid index
            for j in range(_ncells):
                head.data[j] = UINT_MAX

            for j in range(np):
                next.data[j] = UINT_MAX



"""Acoustic wave diffusion in 2-d (2 mins)

Two Dimensional constant pressure accuracy test
particles should simply advect in a periodic domain
"""

# NumPy and standard library imports
import numpy

from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array as gpa
from pysph.solver.application import Application

from pysph.sph.scheme import GasDScheme, ADKEScheme, GSPHScheme, SchemeChooser
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme

# PySPH tools
from pysph.tools import uniform_distribution as ud

# Numerical constants
dim = 2
gamma = 1.4
gamma1 = gamma - 1.0

# solution parameters
dt = 5e-3
tf = 1.


# domain size
xmin = 0.
xmax = 1.
ymin = 0.
ymax = 1.


# scheme constants
alpha1 = 1.0
alpha2 = 0.1
beta = 2.0
kernel_factor = 1.5


class AccuracyTest2D(Application):
    def initialize(self):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.ny = 128
        self.nx = self.ny
        self.dx = (self.xmax - self.xmin) / (self.nx)
        self.hdx = 2.
        self.p = 1.
        self.u = 1
        self.v = -1
        self.c_0 = 1.18
        self.cfl = 0.1

    def add_user_options(self, group):
        group.add_argument(
            "--nparticles", action="store", type=int, dest="nprt", default=256,
            help="Number of particles in domain"
        )

    def consume_user_options(self):
        self.nx = self.options.nprt
        self.ny = self.nx
        self.dx = (self.xmax - self.xmin) / (self.nx)
        self.dt = self.cfl * self.dx / self.c_0

    def create_domain(self):
        return DomainManager(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
            periodic_in_x=True, periodic_in_y=True)

    def create_particles(self):
        global dx
        data = ud.uniform_distribution_cubic2D(
            self.dx, xmin, xmax, ymin, ymax
            )

        x = data[0].ravel()
        y = data[1].ravel()
        dx = data[2]

        volume = dx * dx

        rho = 1 + 0.2 * numpy.sin(
            numpy.pi * (x + y)
        )

        p = numpy.ones_like(x) * self.p

        # const h and mass
        h = numpy.ones_like(x) * self.hdx * dx
        m = numpy.ones_like(x) * volume * rho

        # u = 1
        u = numpy.ones_like(x) * self.u

        # v = -1
        v = numpy.ones_like(x) * self.v

        # thermal energy from the ideal gas EOS
        e = p/(gamma1*rho)

        fluid = gpa(name='fluid', x=x, y=y, rho=rho, p=p, e=e, h=h, m=m,
                    h0=h.copy(), u=u, v=v)
        self.scheme.setup_properties([fluid])

        print("2D Accuracy Test with %d particles"
              % (fluid.get_number_of_particles()))

        return [fluid, ]

    def create_scheme(self):
        self.tf = tf
        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=0, beta=0, k=1.5, eps=0., g1=0., g2=0.,
            has_ghosts=True)

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=kernel_factor, alpha1=0, alpha2=0,
            beta=beta, has_ghosts=True
        )

        crksph = CRKSPHScheme(
            fluids=['fluid'], dim=dim, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=gamma, cl=2, has_ghosts=True
        )

        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.,
            g1=0., g2=0., rsolver=7, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=5.0,
            niter=40, tol=1e-6, has_ghosts=True
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=kernel_factor
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=kernel_factor
        )

        s = SchemeChooser(
            default='gsph', adke=adke, mpm=mpm, gsph=gsph, crksph=crksph,
            psph=psph, tsph=tsph
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'mpm':
            s.configure(kernel_factor=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=True, pfreq=50)
        elif self.options.scheme in ['adke', 'gsph', 'crksph', 'tsph', 'psph']:
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

    def post_process(self):
        from pysph.solver.utils import load
        if len(self.output_files) < 1:
            return
        outfile = self.output_files[-1]
        data = load(outfile)
        pa = data['arrays']['fluid']
        x_c = pa.x
        y_c = pa.y
        rho_c = pa.rho
        rho_e = 1 + 0.2 * numpy.sin(
            numpy.pi * (x_c + y_c)
        )
        num_particles = rho_c.size
        l1_norm = numpy.sum(
            numpy.abs(rho_c - rho_e)
        ) / num_particles

        print(l1_norm)


if __name__ == '__main__':
    app = AccuracyTest2D()
    app.run()
    app.post_process()



from compyle.api import annotate, Elementwise
from compyle.parallel import Scan
from math import floor
from pytools import memoize


def key_func(*args, **kw):
    return args, frozenset(kw.items())


@memoize(key=key_func)
def get_elwise(f, backend):
    return Elementwise(f, backend=backend)


@memoize(key=key_func)
def get_scan(inp_f, out_f, dtype, backend):
    return Scan(input=inp_f, output=out_f, dtype=dtype,
                backend=backend)


@annotate
def exclusive_input(i, ary):
    return ary[i]


@annotate
def exclusive_output(i, prev_item, ary):
    ary[i] = prev_item


@annotate
def norm2(x, y, z):
    return x * x + y * y + z * z


@annotate
def find_cell_id(x, y, z, h, c):
    c[0] = floor((x) / h)
    c[1] = floor((y) / h)
    c[2] = floor((z) / h)


@annotate(p='ulong', return_='ulong')
def interleave1(p):
    return p


@annotate(ulong='p, q', return_='ulong')
def interleave2(p, q):
    p = p & 0xffffffff
    p = (p | (p << 16)) & 0x0000ffff0000ffff
    p = (p | (p << 8)) & 0x00ff00ff00ff00ff
    p = (p | (p << 4)) & 0x0f0f0f0f0f0f0f0f
    p = (p | (p << 2)) & 0x3333333333333333
    p = (p | (p << 1)) & 0x5555555555555555

    q = q & 0xffffffff
    q = (q | (q << 16)) & 0x0000ffff0000ffff
    q = (q | (q << 8)) & 0x00ff00ff00ff00ff
    q = (q | (q << 4)) & 0x0f0f0f0f0f0f0f0f
    q = (q | (q << 2)) & 0x3333333333333333
    q = (q | (q << 1)) & 0x5555555555555555

    return (p | (q << 1))


@annotate(ulong='p, q, r', return_='ulong')
def interleave3(p, q, r):
    p = (p | (p << 32)) & 0x1f00000000ffff
    p = (p | (p << 16)) & 0x1f0000ff0000ff
    p = (p | (p << 8)) & 0x100f00f00f00f00f
    p = (p | (p << 4)) & 0x10c30c30c30c30c3
    p = (p | (p << 2)) & 0x1249249249249249

    q = (q | (q << 32)) & 0x1f00000000ffff
    q = (q | (q << 16)) & 0x1f0000ff0000ff
    q = (q | (q << 8)) & 0x100f00f00f00f00f
    q = (q | (q << 4)) & 0x10c30c30c30c30c3
    q = (q | (q << 2)) & 0x1249249249249249

    r = (r | (r << 32)) & 0x1f00000000ffff
    r = (r | (r << 16)) & 0x1f0000ff0000ff
    r = (r | (r << 8)) & 0x100f00f00f00f00f
    r = (r | (r << 4)) & 0x10c30c30c30c30c3
    r = (r | (r << 2)) & 0x1249249249249249

    return (p | (q << 1) | (r << 2))


@annotate
def find_idx(keys, num_particles, key):
    first = 0
    last = num_particles - 1
    middle = (first + last) / 2

    while first <= last:
        if keys[middle] < key:
            first = middle + 1
        elif keys[middle] > key:
            last = middle - 1
        elif keys[middle] == key:
            if middle == 0:
                return 0
            if keys[middle - 1] != key:
                return middle
            else:
                last = middle - 1
        middle = (first + last) / 2

    return -1


@annotate
def neighbor_boxes(c_x, c_y, c_z, nbr_boxes):
    nbr_boxes_length = 1

    nbr_boxes[0] = interleave3(c_x, c_y, c_z)

    key = declare('ulong')
    for j in range(-1, 2):
        for k in range(-1, 2):
            for m in range(-1, 2):
                if (j != 0 or k != 0 or m != 0) and c_x + m >= 0 and c_y + k >= 0 and c_z + j >= 0:
                    key = interleave3(c_x + m, c_y + k, c_z + j)
                    nbr_boxes[nbr_boxes_length] = key
                    nbr_boxes_length += 1

    return nbr_boxes_length
